/*
 * Copyright (C) 1996,1997,1998
 *	SHUDO Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

package NET.shudo.moba.lang;

import java.util.Random;
import NET.shudo.moba.reflect.*;
import NET.shudo.moba.io.ThreadWriter;
import NET.shudo.moba.io.ObjectWriter;
import NET.shudo.moba.io.ObjectReader;
// for notification
import NET.shudo.moba.place.MobaPlace;
import NET.shudo.moba.place.PlaceEventNotifier;
// for migration
import NET.shudo.moba.place.ThreadExporter;
import java.io.IOException;
import java.net.InetAddress;


/**
 * stream$B$KN.$9$3$H$N=PMh$k!"0\F02DG=$J%9%l%C%I$N%/%i%9!#(B
 */
public
class MobaThread extends Thread {
  private static MetaClass threadMetaClass = null;
  private static Object threadMetaClassMutex = new Object();

  private static Random hashCodeGenerator =
	new Random(System.currentTimeMillis());

  /**
   * hash value
   */
  private int hashValue;
  /**
   * $B@8@.;~9o!#(B
   */
  private long generatedTime;
  /**
   * $B85$N5o>l=j!#(B
   */
  private PlaceAddress genedPlace = null;
  /**
   * $B85$N5o>l=j$rJV$9!#(B
   */
  public PlaceAddress generatedPlace() { return this.genedPlace; }
  /**
   * $B85$N5o>l=j$r@_Dj$9$k!#(B
   */
  public void generatedPlace(PlaceAddress place) {
    this.genedPlace = place;
  }


  public MobaThread() {
    super();  init();
  }
  public MobaThread(Runnable target) {
    super(target);  init();
  }
  public MobaThread(ThreadGroup group, Runnable target) {
    super(group, target);  init();
  }
  public MobaThread(String name) {
    super(name);  init();
  }
  public MobaThread(ThreadGroup group, String name) {
    super(group, name);  init();
  }
  public MobaThread(ThreadGroup group, Runnable target, String name) {
    super(group, target, name);  init();
  }


  private static final int THREAD_ID_LAW = 10000;
  public int id() {
    return (this.hashCode() & 0x7fffffff) % THREAD_ID_LAW;
  }
  public static int id(Object obj) {
    return (obj.hashCode() & 0x7fffffff) % THREAD_ID_LAW;
  }


  public int hashCode() {
    return this.hashValue;
  }

  /**
   * $B%O%C%7%eCM$r%i%s%@%`$KJQ99$9$k!#(B
   */
  public int changeHashCode() {
    this.hashValue = hashCodeGenerator.nextInt();

    return this.hashValue;
  }


  public String toString() {
    return "Moba" + super.toString() + "@" + Integer.toHexString(hashCode());
  }


  private void init() {
    // $B@8@.;~9o$r5-O?(B
    this.generatedTime = System.currentTimeMillis();

    // $B@8@.85$r5-O?(B
    try {
      generatedPlace(new PlaceAddress(InetAddress.getLocalHost()));
    }
    catch (java.net.UnknownHostException e) {
      // not reached.
    }

    // hashCode() $B$r:n@.(B
    this.changeHashCode();

    // java.lang.Thread $B$N(B MetaClass $B$r<hF@(B
    synchronized(threadMetaClassMutex) {
      if (threadMetaClass == null) {
	try {
	  threadMetaClass = MetaClass.get("java.lang.Thread");
	}
	catch (ClassNotFoundException e) {
	  // not reached.
	  e.printStackTrace();  System.exit(1);
	}
      }
    }
  }


  public void run() {
    try {
      super.run();

      // notify
      PlaceEventNotifier notifier = MobaPlace.eventNotifier();
      if (notifier != null)
	notifier.stoppedNotify(this);
    }
    catch (ThreadDeath er) {
      // killed !
      throw er;
    }
/*
    finally {
    }
*/
  }


  /**
   * $B%9%l%C%I$KBP1~$9$k(B Runnable $B%/%i%9$N%$%s%9%?%s%9$rF@$k!#(B
   */
  public Runnable runnableInstance() {
    MetaField field = null;
    try {
      field = threadMetaClass.field("target");
    }
    catch (FieldNotFoundException e) { return null; }

    Runnable ret = (Runnable)field.getObjectValue(this);
    if (ret != null)
      return ret;
    else
      return this;
  }


  /**
   * $B;XDj$7$?%9%l%C%I$r!";XDj$7$?(B ThreadGroup $B$K=jB0$5$;$k!#(B
   */
  public static void setThreadGroup(Thread t, ThreadGroup newGroup)
	throws Exception {
    MetaClass mc;
    MetaMethod mm;
    MetaField mf;

    Object[] invokeArgs = new Object[1];
    invokeArgs[0] = t;

    ThreadGroup origGroup = t.getThreadGroup();
    if (origGroup.equals(newGroup))  return;

    // $B85$N(B group $B$+$iH4$1$5$;$k(B
    mc = MetaClass.get("java.lang.ThreadGroup");
    mm = mc.method("remove", "(Ljava/lang/Thread;)V");
    mm.invoke(origGroup, invokeArgs);
	// t.getThreadGroup().remove(t);

    // MobaPlace.threadPool $B$KF~$l$k(B
    mc = MetaClass.get("java.lang.Thread");
    mf = mc.field("group");
    mf.setObjectValue(t, newGroup);
	// t.group = newGroup;

    mc = MetaClass.get("java.lang.ThreadGroup");
    mm = mc.method("add", "(Ljava/lang/Thread;)V");
    mm.invoke(newGroup, invokeArgs);
	// newGroup.add(t);
  }


  /** $B$3$N%9%l%C%I$rB>$N7W;;5!$K0\F0$5$;$k!#(B(vt. move) */
  public void moveTo(PlaceAddress addr) throws IOException {
    ThreadExporter.moveTo(this, addr);
  }
  /** $B$3$N%9%l%C%I$rB>$N7W;;5!$K0\F0$5$;$k!#(B(vt. move) */
  public void moveTo(String host) throws IOException {
    ThreadExporter.moveTo(this, new PlaceAddress(host));
  }
  /** $B$3$N%9%l%C%I$rB>$N7W;;5!$K0\F0$5$;$k!#(B(vt. move) */
  public void moveTo(InetAddress dest) throws IOException {
    ThreadExporter.moveTo(this, new PlaceAddress(dest));
  }
  /** $B$3$N%9%l%C%I$rB>$N7W;;5!$K0\F0$5$;$k!#(B(vt. move) */
  public void moveTo(String host, int port) throws IOException {
    ThreadExporter.moveTo(this, new PlaceAddress(host, port));
  }
  /** $B$3$N%9%l%C%I$rB>$N7W;;5!$K0\F0$5$;$k!#(B(vt. move) */
  public void moveTo(InetAddress dest, int port) throws IOException {
    ThreadExporter.moveTo(this, new PlaceAddress(dest, port));
  }

  /** $B<+?H(B (Thread#currentThread()) $B$,(B host $B$K0\F0$9$k!#(B(vi. go) */
  public static void goTo(PlaceAddress addr) throws IOException {
    ThreadExporter.moveTo((MobaThread)Thread.currentThread(), addr);
  }
  /** $B<+?H(B (Thread#currentThread()) $B$,(B host $B$K0\F0$9$k!#(B(vi. go) */
  public static void goTo(String host) throws IOException {
    ThreadExporter.moveTo((MobaThread)Thread.currentThread(),
			new PlaceAddress(host));
  }
  /** $B<+?H(B (Thread#currentThread()) $B$,(B host $B$K0\F0$9$k!#(B(vi. go) */
  public static void goTo(InetAddress dest) throws IOException {
    ThreadExporter.moveTo((MobaThread)Thread.currentThread(),
			new PlaceAddress(dest));
  }
  /** $B<+?H(B (Thread#currentThread()) $B$,(B host $B$K0\F0$9$k!#(B(vi. go) */
  public static void goTo(String host, int port) throws IOException {
    ThreadExporter.moveTo((MobaThread)Thread.currentThread(),
			new PlaceAddress(host, port));
  }
  /** $B<+?H(B (Thread#currentThread()) $B$,(B host $B$K0\F0$9$k!#(B(vi. go) */
  public static void goTo(InetAddress dest, int port) throws IOException {
    ThreadExporter.moveTo((MobaThread)Thread.currentThread(),
			new PlaceAddress(dest, port));
  }

  /** $B$3$N%9%l%C%I$rB>$N7W;;5!$K%3%T!<$9$k!#(B(cf. fork(2)) */
  public void forkThisThread(PlaceAddress addr) throws IOException {
    ThreadExporter.fork(this, addr);
  }
  /** $B$3$N%9%l%C%I$rB>$N7W;;5!$K%3%T!<$9$k!#(B(cf. fork(2)) */
  public void forkThisThread(String host) throws IOException {
    ThreadExporter.fork(this, new PlaceAddress(host));
  }
  /** $B$3$N%9%l%C%I$rB>$N7W;;5!$K%3%T!<$9$k!#(B(cf. fork(2)) */
  public void forkThisThread(InetAddress dest) throws IOException {
    ThreadExporter.fork(this, new PlaceAddress(dest));
  }
  /** $B$3$N%9%l%C%I$rB>$N7W;;5!$K%3%T!<$9$k!#(B(cf. fork(2)) */
  public void forkThisThread(String host, int port) throws IOException {
    ThreadExporter.fork(this, new PlaceAddress(host, port));
  }
  /** $B$3$N%9%l%C%I$rB>$N7W;;5!$K%3%T!<$9$k!#(B(cf. fork(2)) */
  public void forkThisThread(InetAddress dest, int port) throws IOException {
    ThreadExporter.fork(this, new PlaceAddress(dest, port));
  }

  /**
   * $B<+?H$rB>$N7W;;5!$K%3%T!<$9$k!#(B(cf. fork(2))
   * @return true to created thread, false to original thread
   */
  public static boolean fork(PlaceAddress addr) throws IOException {
    return ThreadExporter.fork((MobaThread)Thread.currentThread(), addr);
  }
  /**
   * $B<+?H$rB>$N7W;;5!$K%3%T!<$9$k!#(B(cf. fork(2))
   * @return true to created thread, false to original thread
   */
  public static boolean fork(String host) throws IOException {
    return ThreadExporter.fork((MobaThread)Thread.currentThread(),
			new PlaceAddress(host));
  }
  /**
   * $B<+?H$rB>$N7W;;5!$K%3%T!<$9$k!#(B(cf. fork(2))
   * @return true to created thread, false to original thread
   */
  public static boolean fork(InetAddress dest) throws IOException {
    return ThreadExporter.fork((MobaThread)Thread.currentThread(),
			new PlaceAddress(dest));
  }
  /**
   * $B<+?H$rB>$N7W;;5!$K%3%T!<$9$k!#(B(cf. fork(2))
   * @return true to created thread, false to original thread
   */
  public static boolean fork(String host, int port) throws IOException {
    return ThreadExporter.fork((MobaThread)Thread.currentThread(),
			new PlaceAddress(host, port));
  }
  /**
   * $B<+?H$rB>$N7W;;5!$K%3%T!<$9$k!#(B(cf. fork(2))
   * @return true to created thread, false to original thread
   */
  public static boolean fork(InetAddress dest, int port) throws IOException {
    return ThreadExporter.fork((MobaThread)Thread.currentThread(),
			new PlaceAddress(dest, port));
  }


  private boolean internalized = false;

  /**
   * $B$3$N%9%l%C%I$,(B internalization $B$G9=@.$5$l$?$b$N$+D4$Y$k!#(B
   *
   * @return internalized $B%9%l%C%I$J$i??$rJV$9!#(B
   * @see NET.shudo.moba.io.ObjectReader#readObject()
   */
  public boolean isInternalized() { return internalized; }

  public boolean isInternalized(boolean flag) {
    boolean ret = internalized;	// $BJ]B8(B
    internalized = flag;
    return ret;
  }


  /**
   * internalize $B$5$l$?(B MobaThread $B$r5/F0$9$k!#(BThread#start() $B$NBeBX(B<BR>
   * boot() $B$7$?%9%l%C%I$O(B Thread#resume() $B$G<B9T$r:F3+$9$k!#(B<BR>
   * MOBA place $B$K0\F0$7$?%9%l%C%I$O<+F0E*$K(B boot(), resume() $B$5$l$k!#(B
   *
   * @see java.lang.Thread#start()
   * @see java.lang.Thread#resume()
   */
  public native void boot();

  /**
   * $BM?$($i$l$?%9%l%C%I$N(B context $B$rM?$($i$l$?%9%H%j!<%`$K=q$-=P$9!#(B
   */
  public native void writeContextTo(ObjectWriter out) throws IOException;

  /**
   * $BM?$($i$l$?%9%l%C%I$N(B context $B$rM?$($i$l$?%9%H%j!<%`$+$i9=@.$9$k!#(B
   */
  public static native void readContextFrom(
		ObjectReader in, Thread thr, ClassLoader loader);


  /**
   * $B$3$N%9%l%C%I$N8=:_$N(B PC $B$K4X$9$k>pJs$rJ8;zNs$H$7$FF@$k!#(B
   */
  public native String currentPCString();


  // $B%i%$%V%i%j$N07$$(B
  private static boolean libLoaded = false;
  private static Object libLoadedLock = new Object();

  /**
   * native methods $B$r4^$`%i%$%V%i%j$r%m!<%I$9$k!#(B
   */
  public static void loadLibrary() {
    synchronized(libLoadedLock) {
      if (!libLoaded) {
	// native method $B%i%$%V%i%j$N%m!<%I(B
	System.loadLibrary("Moba");
	libLoaded = true;
      }
    }
  }
}
