/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

package NET.shudo.moba.lang;


import java.net.InetAddress;
import java.net.UnknownHostException;
// $B<+A0$N30It2=$N$?$a(B
import java.io.IOException;
import NET.shudo.moba.io.Streamable;
import NET.shudo.moba.io.ObjectWriter;
import NET.shudo.moba.io.ObjectReader;


/**
 * Place $B$N%"%I%l%9!#(B<BR>
 * Place $B$N0LCV$O(B IP $B%"%I%l%9$H%]!<%HHV9f$NAH$GI=8=$5$l$k!#(B
 *
 * @see NET.shudo.moba.place.MobaPlace
 */
public class PlaceAddress implements Streamable {
  /**
   * MOBA place $B$N%]!<%HHV9f!#(B
   */
  public static int PLACE_PORT = 10000;

  /**
   * CommandServer $B$,BT$D%]!<%HHV9f$N!"(Bplace $B$N%]!<%HHV9f$KBP$9$k:9J,!#(B
   */
  public static final int COMMAND_PORT_DIFF = 0;
  /**
   * ClassServer $B$,BT$D%]!<%HHV9f$N!"(Bplace $B$N%]!<%HHV9f$KBP$9$k:9J,!#(B
   */
  public static final int CLASS_PORT_DIFF = 1;
  /**
   * ThreadServer $B$,BT$D%]!<%HHV9f$N!"(Bplace $B$N%]!<%HHV9f$KBP$9$k:9J,!#(B
   */
  public static final int THREAD_PORT_DIFF = 2;

  private static int controllerPort = PLACE_PORT + 10;	// 10010/tcp
  private static int controllerPortDiff = 10;
  public static void controllerPort(int port) {
    controllerPort = port;
    controllerPortDiff = port - PLACE_PORT;
  }
  public static int controllerPort() {
    return controllerPort;
  }
  public static int controllerPortDiff() {
    return controllerPortDiff;
  }


  private static InetAddress localInetAddress;

  static {
    try {
      localInetAddress = InetAddress.getLocalHost();
    }
    catch (UnknownHostException e) {
      System.err.println("Cannot get local IP address.");
      System.exit(1);
    }
  }


  private InetAddress IPAddr;
  private int portno = PLACE_PORT;


  public PlaceAddress() {}

  public PlaceAddress(InetAddress addr, int portno) {
    this.IPAddr = addr;
    this.portno = portno;
  }
  public PlaceAddress(InetAddress addr) {
    this(addr, PLACE_PORT);
  }

  public PlaceAddress(String host, int portno) throws UnknownHostException {
    this.IPAddr = InetAddress.getByName(host);
    this.portno = portno;
  }
  public PlaceAddress(String host) throws UnknownHostException {
	// accept hostpart:port style
    String hostpart = host;
    int port = PLACE_PORT;

    // hostpart:port $B7A<0$KBP1~(B
    int colonIndex = host.lastIndexOf(':');
    if (colonIndex >= 0) {
      hostpart = host.substring(0, colonIndex);
      String portpart = host.substring(colonIndex + 1);
      try { port = Integer.parseInt(portpart); }
      catch (NumberFormatException e) { port = PLACE_PORT; }
    }

    // hostname/ipaddr $B7A<0$KBP1~(B
    int slashIndex = host.lastIndexOf('/');
    if (slashIndex >= 0) {
      hostpart = host.substring(0, slashIndex);
    }

    this.IPAddr = InetAddress.getByName(hostpart);
    this.portno = port;
  }


  public int hashCode() {
    return IPAddr.hashCode() + portno * 65535;	// to be improved
  }

  public boolean equals(Object obj) {
    if (obj instanceof PlaceAddress) {
      PlaceAddress addr = (PlaceAddress)obj;
      if (this.IPAddr.equals(addr.IPAddr) && (this.portno == addr.portno))
	return true;
    }
    return false;
  }

  public String toString() {
//    return IPAddr.getHostName() + ":" + portno;
    return IPAddr.getHostAddress() + ":" + portno;
  }


  public int port() { return this.portno; }
  public InetAddress inetAddress() { return this.IPAddr; }

  public String hostName() { return this.IPAddr.getHostName(); }
  public String hostAddress() { return this.IPAddr.getHostAddress(); }


  /**
   * $B%m!<%+%k%"%I%l%9$rJV$9!#(B
   */
  public static PlaceAddress localAddress() {
    return localAddress;
  }

  private static PlaceAddress localAddress =
			new PlaceAddress(localInetAddress, PLACE_PORT);

  /**
   * $B$3$N%"%I%l%9$,%m!<%+%k$J$i??$rJV$9!#(B
   */
  public boolean isLocalAddress() {
    return this.equals(localAddress);
  }


  /**
   * $B<+A0$N30It2=%a%=%C%I!#(B
   */
  public void writeToStream(ObjectWriter out) throws IOException {
    out.writeUTF(this.IPAddr.getHostAddress() + ":" + this.portno);
  }
  /**
   * $B<+A0$NFbIt2=%a%=%C%I!#(B
   */
  public void readFromStream(ObjectReader in) throws IOException {
    String host = in.readUTF();
    int colonIndex = host.lastIndexOf(':');

    String hostpart = host.substring(0, colonIndex);
    this.IPAddr = InetAddress.getByName(hostpart);

    String portpart = host.substring(colonIndex + 1);
    try { this.portno = Integer.parseInt(portpart); }
    catch (NumberFormatException e) { this.portno = PLACE_PORT; }
  }
}
