/*
 * Copyright (C) 1996,1997,1998,1999,2003
 *	Kazuyuki Shudo
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

/* native methods for MetaClass */

#include <stdio.h>	/* for NULL */

#include "NET_shudo_moba_reflect_MetaClass.h"

#include "moba.h"


/*
 * $B%*%V%8%'%/%H(B ($BG[Ns$r=|$/(B) $B$N%$%s%9%?%s%9$r:n$k!#(B
 * $BG[Ns$K$D$$$F$O(B MetaClass#newArray() $B$rMQ$$$k!#(B
 */
JNIEXPORT jobject JNICALL Java_NET_shudo_moba_reflect_MetaClass_newInstance
  (JNIEnv *env, jclass _this, jclass clazz) {
  jobject obj;
  jmethodID constructorID = NULL;

  /* create an instance */
  obj = (*env)->AllocObject(env, clazz);
  if ((*env)->ExceptionOccurred(env)) {
#ifdef DEBUG
    printf("MetaClass#newInstance(): AllocObject() threw an exception.\n");
    fflush(stdout);
#endif
    return NULL;
  }

  /* search constructor */
  /* superclass $B$rC)$j0z?t$J$7$N%3%s%9%H%i%/%?$rC5$9(B */
  do {
    if (constructorID = (*env)->GetMethodID(env, clazz, "<init>", "()V"))
      break;
  } while (clazz = (*env)->GetSuperclass(env, clazz));

  /* call the constructor */
  if (constructorID) {
#if 1
    (*env)->CallObjectMethod(env, obj, constructorID);
#else
    callMethod(env, obj, constructorID, 0);
#endif
  }
  else {
#ifdef DEBUG
    printf("MetaClass#newInstance(): constructor is not found.\n");
    fflush(stdout);
#endif
  }

  return obj;
}


/*
 * $BG[Ns$N%$%s%9%?%s%9$r:n$k!#(B
 * $BG[Ns0J30$N%*%V%8%'%/%H$N@8@.$K$O(B MetaClass#newInstance() $B$rMQ$$$k!#(B
 */
JNIEXPORT jobject JNICALL Java_NET_shudo_moba_reflect_MetaClass_newArray
  (JNIEnv *env, jclass _this, jstring sigStr, jint len, jobject loader) {
  const char *sig = (*env)->GetStringUTFChars(env, sigStr, NULL);

  jarray array = NULL;

  jclass elemClazz;
  char *elemSig;
  char *p;

  if (sig[0] != '[')  return NULL;	/* $BG0$N$?$a(B */
#ifdef DEBUG
  printf("MetaClass#newArray(): signature is `%s'\n", sig);
  fflush(stdout);
#endif

  switch (sig[1]) {
  case 'L':
    elemSig = strdup(sig + 2);
    p = elemSig;
    while (*p != '\0') {
      if (*p == ';') { *p = '\0';  break; }
      p++;
    }
    elemClazz = classLoaderFindClass(env, elemSig, loader);
    free(elemSig);
    array = (*env)->NewObjectArray(env, (jsize)len, elemClazz, NULL);
    break;
  case '[':
    elemClazz = classLoaderFindClass(env, sig + 1, loader);
    array = (*env)->NewObjectArray(env, (jsize)len, elemClazz, NULL);
    break;
  case 'Z':
    array = (*env)->NewBooleanArray(env, (jsize)len);
    break;
  case 'B':
    array = (*env)->NewByteArray(env, (jsize)len);
    break;
  case 'C':
    array = (*env)->NewCharArray(env, (jsize)len);
    break;
  case 'S':
    array = (*env)->NewShortArray(env, (jsize)len);
    break;
  case 'I':
    array = (*env)->NewIntArray(env, (jsize)len);
    break;
  case 'J':
    array = (*env)->NewLongArray(env, (jsize)len);
    break;
  case 'F':
    array = (*env)->NewFloatArray(env, (jsize)len);
    break;
  case 'D':
    array = (*env)->NewDoubleArray(env, (jsize)len);
    break;
  default:
#ifdef DEBUG
    printf("MetaClass#newArray(): invalid signature: %s\n", sig);
    fflush(stdout);
#endif
    array = NULL;
    break;
  }

  (*env)->ReleaseStringUTFChars(env, sigStr, sig);

  return (jobject)array;
}


/*
 * $B%/%i%9$rDj5A$9$k!#(B(cf. ClassLoader#defineClass())
 * $BCm(B) MetaClass $B$K$"$k$Y$-%a%=%C%I$G$O$J$$!#(B
 */
JNIEXPORT jclass JNICALL Java_NET_shudo_moba_reflect_MetaClass_defineClass
  (JNIEnv *env, jclass _this, jobject loader, jstring nameStr, jbyteArray clazzData, jint offset, jint length) {
  const char *name = (*env)->GetStringUTFChars(env, nameStr, NULL);

  jbyte *data = (*env)->GetByteArrayElements(env, clazzData, NULL);
  jclass clazz =
	(*env)->DefineClass(env, name, loader, data + offset, (jsize)length);
  (*env)->ReleaseByteArrayElements(env, clazzData, data, JNI_ABORT);

  (*env)->ReleaseStringUTFChars(env, nameStr, name);

  return clazz;
}


/*
 * $B<+A0$N(B writeObject() $B$r;}$C$F$$$k$+JV$9!#(B
 */
JNIEXPORT jboolean JNICALL Java_NET_shudo_moba_reflect_MetaClass_hasOwnWriteObject0
  (JNIEnv *env, jobject _this) {
  /* MetaClass#thisClass $B$rF@$k(B */
  jfieldID fid = (*env)->GetFieldID(env, (*env)->GetObjectClass(env, _this),
	"thisClass", "Ljava/lang/Class;");
  jclass clazz = (jclass)(*env)->GetObjectField(env, _this, fid);

  jmethodID mid = (*env)->GetMethodID(env, clazz,
	OWN_WRITEOBJ_METHOD, OWN_WRITEOBJ_SIGNATURE);
  if ((*env)->ExceptionOccurred(env)) {
    (*env)->ExceptionClear(env);
    mid = NULL;
  }

  return (mid ? JNI_TRUE : JNI_FALSE);
}

/*
 * $B<+A0$N(B readObject() $B$r;}$C$F$$$k$+JV$9!#(B
 */
JNIEXPORT jboolean JNICALL Java_NET_shudo_moba_reflect_MetaClass_hasOwnReadObject0
  (JNIEnv *env, jobject _this) {
  /* MetaClass#thisClass $B$rF@$k(B */
  jfieldID fid = (*env)->GetFieldID(env, (*env)->GetObjectClass(env, _this),
	"thisClass", "Ljava/lang/Class;");
  jclass clazz = (jclass)(*env)->GetObjectField(env, _this, fid);

  jmethodID mid = (*env)->GetMethodID(env, clazz,
	OWN_READOBJ_METHOD, OWN_READOBJ_SIGNATURE);
  if ((*env)->ExceptionOccurred(env)) {
    (*env)->ExceptionClear(env);
    mid = NULL;
  }
#if 0
  printf("MetaClass#hasOwnReadObject0: return %s.\n", (mid ? "true" : "false"));
  fflush(stdout);
#endif

  return (mid ? JNI_TRUE : JNI_FALSE);
}
