/*
 * Copyright (C) 1996,1997,1998,1999,2000
 *	SHUDO Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

/* native methods for thread streaming */

#include <stdio.h>	/* for NULL */
#include <string.h>	/* for memcpy() */
#include <sys/types.h>	/* for open(),... */
#include <sys/stat.h>	/* for open(),... */
#include <fcntl.h>	/* for open(),... */
#include <unistd.h>	/* for sleep() */

#include "NET_shudo_moba_lang_MobaThread.h"

#include <jni.h>	/* for native methods */

#include <native.h>	/* for native methods */
#if 0
#include <sys_api.h>	/* for type sys_thread_t, sys*() */
#else
typedef struct sys_thread  sys_thread_t;
typedef void *  stackp_t;
#endif

#if 0
#include "threads.h"	/* for JavaStackSize, THR_SYSTEM, THR_USER */
#else
#include "java_lang_Thread0.h"	/* for struct Classjava_lang_Thread */
#include "java_lang_ThreadGroup0.h"
#define THR_SYSTEM	0
#define THR_USER	1
#define threadSetPriority(hThread, i) \
	sysThreadSetPriority((sys_thread_t *)(unhand(hThread)->PrivateInfo), i)
#define threadInit(hThread, sb) \
	sysThreadInit((sys_thread_t *)(unhand(hThread)->PrivateInfo), sb)
#define threadExit()	sysThreadExit()
extern long JavaStackSize;
extern long ProcStackSize;
#endif

#include "moba.h"

#include "stack.h"


/*
 * Variables
 */
#define EXCEPTION_BUF_SIZE	100
static char exception_msg_buf[EXCEPTION_BUF_SIZE];


/*
 * Local Functions
 */
static void MobaThreadRT(Hjava_lang_Thread *p);
static unsigned char *returnPC(JavaFrame *frame);

static JavaFrame *allocNewJavaFrame(struct methodblock *mb, ExecEnv *ee);
static JavaFrame *allocNewJavaFrameWithNLocals(
			struct methodblock *mb, ExecEnv *ee, int nlocals);

static int writeClassAndMethod(JNIEnv *env,
		jobject /*OutputStream*/ out, struct methodblock *method);
static struct methodblock *readClassAndMethod(JNIEnv *env,
		jobject /*InputStream*/ in, jobject /*ClassLoader*/ loader);

static int writeStackItem(JNIEnv *env,
	jobject /*ObjectWriter*/ out, stack_item *itemptr, char *sig);
static int writeStackItemAuto(JNIEnv *env,
	jobject /*ObjectWriter*/ out, stack_item *itemptr);
static int readStackItem(JNIEnv *env,
	jobject /*ObjectReader*/ in, stack_item *dest);

static void dumpAllFrames(FILE *out, JavaFrame *topFrame);
static void dumpFrame(FILE *out, JavaFrame *frame);


/*
 * serialize the internal of the thread.
 */
JNIEXPORT void JNICALL Java_NET_shudo_moba_lang_MobaThread_writeContextTo
  (JNIEnv *env, jobject _this, jobject /*ObjectWriter*/ out) {
  Hjava_lang_Thread *thisHandle = (Hjava_lang_Thread *)DeRef(env, _this);
  ExecEnv *ee = (ExecEnv *)unhand(thisHandle)->eetop;
  JavaFrame *frame;
  struct methodblock *method;

  uint16_t nFrames;
  uint32_t lastpc_offset;


#ifdef DEBUG
  printf("writeContextTo() called.\n");
#endif

  /* JavaFrame $B$r?7$7$$=g$K%9%?%C%/$K@Q$`(B */
  stack_init();
  nFrames = 0;
  frame = ee->current_frame;
#if 0
  /* $B$3$N%3!<%I$rM-8z$K$9$k$H(B examples/ThreadWriterTest.java $B$,F0$+$J$$!D(B
	Oct 6 1997 */
  dumpAllFrames(stdout, frame);
  fflush(stdout);
#endif
  while (frame) {	/* assume initial_frame->prev == NULL */
    if (frame->current_method) {
      nFrames++;
      stack_push((void *)frame);
    }
    else {
      /* current_method $B$r;}$?$J$$(B frame
	 $B$D$^$j(B main $B%a%=%C%I$OL5;k(B */
      break;
    }
    frame = frame->prev;
  }

#ifdef DEBUG
  printf("# of frames: %d\n", (int)nFrames);
  fflush(stdout);
#endif
  if ((writeInt16(env, out, (jshort *)&nFrames)) < 0)
    goto exc_occurred;


  /* $B8E$$(B JavaFrame $B$+$iC)$k(B */
  while (!stack_is_empty()) {
    int frameNo = nFrames - stack_depth();	/* for debug */

    frame = (JavaFrame *)stack_pop();
    method = frame->current_method;
#ifdef DEBUG
    printf("\n");
    printf("writing JavaFrame (%d)\n", frameNo);
    fflush(stdout);
#endif

    /* JavaFrame ($B$NO":?(B) $B$r=q$-=P$9(B */
#ifdef EASY_READING
    {
      uint16_t frameNo16 = (uint16_t)frameNo;
      if ((writeChar(env, out, '{')) < 0)  goto exc_occurred;
      if ((writeInt16(env, out, (jshort *)&frameNo16)) < 0)  goto exc_occurred;
      if ((writeChar(env, out, '}')) < 0)  goto exc_occurred;
    }
#endif

    /* $B%/%i%9!"%a%=%C%I(B */
    {
      uint16_t methodAcc = method->fb.access;

      if ((writeClassAndMethod(env, out, method)) < 0)  goto exc_occurred;
      if ((writeInt16(env, out, (jshort *)&methodAcc)) < 0)  goto exc_occurred;
	/* access flag */
    }
    /* $BJQ0L(B or $B$=$N$^$^$NCM$G=q$-=P$5$l$k$b$N(B */
    /* returnpc, optop, lastpc, mon_starttime */
    {
      uint32_t returnpc_offset;		/* 4 bytes */
      int16_t optop_offset;		/* 2 bytes $B$G=<J,(B (< 2000) */
      /* uint32_t lastpc_offset;		/* 4 bytes */

      returnpc_offset = pc2offset(method, frame->returnpc);
      optop_offset = sp2offset(frame, frame->optop);
      lastpc_offset = pc2offset(method, frame->lastpc);
#ifdef DEBUG
      printf("returnpc: %d\n", returnpc_offset);
      printf("optop:    %d\n", (int)optop_offset);
      printf("lastpc:   %d\n", lastpc_offset);
      fflush(stdout);
#endif

      if ((writeInt32(env, out, (jint *)&returnpc_offset)) < 0)
	goto exc_occurred;
      if ((writeInt16(env, out, (jshort *)&optop_offset)) < 0)
	goto exc_occurred;
      if ((writeInt32(env, out, (jint *)&lastpc_offset)) < 0)
	goto exc_occurred;

#if 0
      /* JavaFrame.mon_starttime */
      if ((writeInt32(env, out, (jint *)&frame->mon_starttime)) < 0)
	goto exc_occurred;
#endif
    }
    /* constant_pool */
    /* $B7?$,(B returnpc $B$K0MB8$9$k(B
     *  returnpc is 0: (struct methodblock *)
     *       is not 0: (cp_item *)
     *  ExecuteJava() (in javasrc/share/java/runtime/executeJava.c) $B$h$j(B
     */
    {
#if (JDK_VER < 110)
      if ((writeChar(env, out, CP_CP)) < 0)  goto exc_occurred;
#else
      if (frame->returnpc == 0) {
	/* (struct methodblock *)frame->constant_pool */
	struct methodblock *targetMethod =
		(struct methodblock *)frame->constant_pool;
	if ((writeChar(env, out, CP_METHOD)) < 0)  goto exc_occurred;

	if ((writeClassAndMethod(env, out, targetMethod)) < 0)  goto exc_occurred;
      }
      else {
	if ((writeChar(env, out, CP_CP)) < 0)  goto exc_occurred;
      }
#endif /* JDK_VER < 110 */
    }
    /* local variables */
#ifdef DEBUG
    printf("writing local vars.\n");
    printf(" (#localvar: %d, nlocals: %d)\n",
	LocalVarTblLen(method), method->nlocals);
    fflush(stdout);
#endif
    {
      long slot;  uint16_t slot16;
      int i;

      int lvTblLen = LocalVarTblLen(method);
      if (lvTblLen > 0) {	/* the method has a LocalVariableTable attr. */
	char *sig;
  #ifdef DEBUG
	char *name;
	long pc0, length;
  #endif

	for (i = 0; i < LocalVarTblLen(method); i++) {
	  slot = LocalVarSlot(method, i);
	  sig = LocalVarSig(method, i);
  #ifdef DEBUG
	  name = LocalVarName(method, i);
	  pc0 = LocalVarStartPC(method, i);
	  length = LocalVarLengthPC(method, i);
	  printf(" [%d] %s %s\tslot:%ld\tPC:%ld(%ld)\n",
		i, sig, name, slot, pc0, length);
	  fflush(stdout);
  #endif
	  if (isLocalVarAlive(method, i, lastpc_offset)) {
	    /* local var. is alive */
  #ifdef DEBUG
	    printf("    is alive.\n");  fflush(stdout);
  #endif
	    /* slot */
	    slot16 = (uint16_t)slot;
  #ifdef EASY_READING
	    if ((writeChar(env, out, '[')) < 0)  goto exc_occurred;
  #endif
	    if ((writeInt16(env, out, (jshort *)&slot16)) < 0)
	      goto exc_occurred;
  #ifdef EASY_READING
	    if ((writeChar(env, out, ']')) < 0)  goto exc_occurred;
  #endif
	    /* content */
	    if ((writeStackItem(env, out, frame->vars + slot, sig)) < 0)
	      goto exc_occurred;
	  }	/* local var. is alive. */
	}
      }
      else {	/* the method doesn't have  a LocalVariableTable attr. */
	for (slot = 0; slot < method->nlocals; slot++) {
  #ifdef DEBUG
	  printf(" slot:%ld\n", slot);
	  fflush(stdout);
  #endif
	  /* slot */
	  slot16 = (uint16_t)slot;
  #ifdef EASY_READING
	  if ((writeChar(env, out, '[')) < 0)  goto exc_occurred;
  #endif
	  if ((writeInt16(env, out, (jshort *)&slot16)) < 0)  goto exc_occurred;
  #ifdef EASY_READING
	  if ((writeChar(env, out, ']')) < 0)  goto exc_occurred;
  #endif
	  /* content */
	  if ((writeStackItemAuto(env, out, frame->vars + slot)) < 0)
	    goto exc_occurred;
	}
      }		/* if (lvTblLen > 0) */


      {
	/* terminate local vars. */
	uint16_t term = 0xffff;
#ifdef EASY_READING
	if ((writeChar(env, out, '[')) < 0)  goto exc_occurred;
#endif
	if ((writeInt16(env, out, (jshort *)&term)) < 0)
	  goto exc_occurred;
#ifdef EASY_READING
	if ((writeChar(env, out, ']')) < 0)  goto exc_occurred;
#endif
      }
    }
    /* stack, JavaFrame.optop */
    {
      stack_item *item = frame->ostack;
#ifdef DEBUG
      int i = 0;
      if (item < frame->optop) {
	printf("writing stack items: %d\n", frame->optop - item);
	fflush(stdout);
      }
#endif
      while (item < frame->optop) {
#ifdef DEBUG
	printf("writing frame->ostack[%d]\n", i++);
	fflush(stdout);
#endif
	if ((writeStackItemAuto(env, out, item)) < 0)  goto exc_occurred;
	item++;
      }
    }
  }	/* follow a chain of JavaFrames */

#ifdef DEBUG
  printf("writeContextTo() done.\n");
  fflush(stdout);
#endif

  return;

exc_occurred:
  {
    jclass exc_clazz = (*env)->FindClass(env, "java/io/IOException");
    (*env)->ThrowNew(env, exc_clazz,
	"MobaThread#writeContextTo: write failed.");
#ifdef DEBUG
    printf("writeContextTo(): exc. occurred.\n");
    fflush(stdout);
#endif
  }
}


/*
 * deserialize the internal of the thread.
 */
JNIEXPORT void JNICALL Java_NET_shudo_moba_lang_MobaThread_readContextFrom
  (JNIEnv *env, jclass clazz,
	jobject /*ObjectReader*/ in, jobject /*Thread*/ obj,
	jobject /*ClassLoader*/ loader) {
  ExecEnv *ee = JNIEnv2EE(env);
  Hjava_lang_Thread *objHandle = (Hjava_lang_Thread *)DeRef(env, obj);

  ExecEnv *targetEE;
  JavaFrame *frame;
  struct methodblock *method;

  uint16_t nFrames;
  int frameCount;

#ifdef DEBUG
  printf("readContextFrom() is called.\n");
  printf("objHandle: 0x%x\n", objHandle);
  fflush(stdout);
#endif
  targetEE = (ExecEnv *)malloc(sizeof(ExecEnv));

  InitializeExecEnv(targetEE, (JHandle *)NULL);
  /*
   * cannot call as `InitializeExecEnv(targetEE, (JHandle *)objHandle)'.
   * because InitializeExecEnv() expects
   * unhand(<the thread>)->initiali_stack_memory is already initialized.
   */
  unhand(objHandle)->eetop = (int32_t)targetEE;
  targetEE->thread = (JHandle *)objHandle;

#if 0
  /* fake for JDK profiling */
  /* allocate null frame */
  initialNullFrame = allocNewJavaFrame(NULL, targetEE);
#endif


  nFrames = (uint16_t)readInt16(env, in);
#ifdef DEBUG
  printf("# of frames: %d\n", (int)nFrames);
  fflush(stdout);
#endif


  for (frameCount = 0; frameCount < nFrames; frameCount++) {
#ifdef DEBUG
    printf("\n");
    printf("reading stack frame(%d):\n", frameCount);
    fflush(stdout);
#endif
    /* reconstruct the chain of JavaFrames */
#ifdef EASY_READING
    {
      uint16_t frameNo16 = -1;
      readChar(env, in);
      frameNo16 = (uint16_t)readInt16(env, in);
      readChar(env, in);
    }
#endif

    /* class and method name */
    {
      uint16_t methodAcc;
      int i;

      method = readClassAndMethod(env, in, loader);
      methodAcc = (uint16_t)readInt16(env, in);	/* access flag */

      if (!method)  return;


      /* check properness of the access flag */
#define ACC_EQ_MASK	(ACC_NATIVE | ACC_MACHINE_COMPILED)
      if ((methodAcc & ACC_EQ_MASK) != (method->fb.access & ACC_EQ_MASK)) {
	jio_snprintf(exception_msg_buf, EXCEPTION_BUF_SIZE,
		"%s#%s", cbName(method->fb.clazz), method->fb.name);
	SignalError(0, MOBA_PACKAGE "FieldNotCompatibleException",
		exception_msg_buf);
	return;
      }
    } /* class and method name */


    /* allocate a new JavaFrame */
    frame = allocNewJavaFrame(method, targetEE);

    /* written in offset or immediate value: */
    /*   returnpc, optop, lastpc, mon_starttime */
    /* constant_pool is already set in allocNewJavaFrame(). */
    {
      uint32_t returnpc_offset;		/* 4 bytes */
      int16_t optop_offset;		/* 2 bytes */
      uint32_t lastpc_offset;		/* 4 bytes */

      returnpc_offset = (uint32_t)readInt32(env, in);
      optop_offset = readInt16(env, in);
      lastpc_offset = (uint32_t)readInt32(env, in);
#ifdef DEBUG
      printf("returnpc: %d\n", returnpc_offset);
      printf("optop:    %d\n", (int)optop_offset);
      printf("lastpc:   %d\n", lastpc_offset);
      fflush(stdout);
#endif

      frame->returnpc = offset2pc(method, returnpc_offset);
      frame->optop = offset2sp(frame, optop_offset);
      frame->lastpc = offset2pc(method, lastpc_offset);

#if 0
      /* JavaFrame.mon_starttime */
      frame->mon_starttime = (int)readInt32(env, in);
#endif
    }
    /* constant_pool */
    {
      struct methodblock *method;
      char cpType = readChar(env, in);

#if (JDK_VER >= 110)
      switch(cpType) {
      case CP_CP:
	/* not to be set, because already set in allocNewJavaFrame() */
	break;
      case CP_METHOD:
	method = readClassAndMethod(env, in, loader);
	frame->constant_pool = (cp_item_type *)method;
#ifdef DEBUG
	printf("constant_pool is pointing method: 0x%x (%s#%s)\n",
		(int)method, cbName(method->fb.clazz), method->fb.name);
#endif
	break;
      default:
	SignalError(0, MOBA_PACKAGE "StreamFormatException",
			"cp type is expected.");
	return;
	break;
      }
#endif /* JDK_VER >= 110 */
    }
    /* local variables */
    {
      uint16_t slot = 0;
      int itemType;

#ifdef DEBUG
      printf("reading local vars.\n");
      fflush(stdout);
#endif
      while (1) {
#ifdef EASY_READING
	readChar(env, in);
#endif
	slot = (uint16_t)readInt16(env, in);
#ifdef EASY_READING
	readChar(env, in);
#endif
	if (slot == (uint16_t)0xffff) {
	  break;
	}
#ifdef DEBUG
	printf("slot:%d\n", (int)slot);
	fflush(stdout);
#endif

	itemType = readStackItem(env, in, &(frame->vars[slot]));
#ifdef DEBUG
	{
	  stack_item item = frame->vars[slot];

	  printf("\tlocal var[%d]: ", (int)slot);
	  switch (itemType) {
	  case STACK_ITEM_HANDLE:
	    if (item.h) {
	      if (obj_flags(item.h) == T_NORMAL_OBJECT)
		printf("%s", cbName(obj_classblock(item.h)));
	      else
		printf("array");
	    }
	    else
	      printf("(null)");
	    break;
	  case STACK_ITEM_IMM32:
	    printf("%d", item.i);
	    break;
	  case STACK_ITEM_IMM64:
	    printf("0x%x, 0x%x", frame->vars[slot].i, frame->vars[slot+1].i);
	    break;
	  default:
	    printf("0x%x", item.i);
	    break;
	  }
	  printf("\n");
	  fflush(stdout);
	}
#endif
      }
    }
    /* stack contents */
    {
      stack_item *stack = frame->ostack;

      while (stack < frame->optop) {
	readStackItem(env, in, stack);
	stack++;
      }
    }

    /* restore members of JavaFrame which not to be saved */
    {
      /* prev */
      /* set in allocNewJavaFrame() */

      /* vars */
      /* set in allocNewJavaFrame() */

      /* javastack */
      /* frame $B$N%a%b%j3NJ];~$K@_Dj$5$l$F$$$k(B */
    }
  }	/* JavaFrame $B$rC)$k(B */


#ifdef DEBUG
  printf("\nExecEnv->current_frame:\n");
  dumpAllFrames(stdout, targetEE->current_frame);
  fflush(stdout);
#endif

#ifdef DEBUG
  printf("readContextFrom() done.\n");
  fflush(stdout);
#endif
}


/**
 * $B8=:_$N(B PC $B$K4X$9$k>pJs$rF@$k!#(B
 */
JNIEXPORT jstring JNICALL Java_NET_shudo_moba_lang_MobaThread_currentPCString
  (JNIEnv *env, jobject _this) {
  Hjava_lang_Thread *thrHandle = (Hjava_lang_Thread *)DeRef(env, _this);
  ExecEnv *ee = (ExecEnv *)((unhand(thrHandle))->eetop);
  JavaFrame *frame = ee->current_frame;
  struct methodblock *method = frame->current_method;
  unsigned char *pc = frame->lastpc;

#define PC_BUF_SIZE 100
  static char buf[PC_BUF_SIZE];

  pc2string(pc, method, buf, buf + PC_BUF_SIZE);

  return (*env)->NewStringUTF(env, buf);
}


/*
 * internalize $B$5$l$?(B MobaThread $B$r5/F0$9$k!#(B
 */
JNIEXPORT void JNICALL Java_NET_shudo_moba_lang_MobaThread_boot
  (JNIEnv *env, jobject _this) {
  Hjava_lang_Thread *thrHandle;
  struct Classjava_lang_Thread *tid;

  (*env)->MonitorEnter(env, _this);

  thrHandle = DeRef(env, _this);
  tid = (struct Classjava_lang_Thread *)unhand(thrHandle);

  if (tid->PrivateInfo) {
    SignalError(0, "java/lang/IllegalThreadStateException", 0);
  }
  else {
    int flag = (tid->daemon ? THR_SYSTEM : THR_USER);
    if (threadCreate(thrHandle, flag, ProcStackSize,
		(void *(*)(void *))MobaThreadRT)) {
      SignalError(0, "java/lang/OutOfMemoryError", 0);
    }
    else {
      threadSetPriority(thrHandle, tid->priority);
      /* threadResume(thread); */
    }
  }

  (*env)->MonitorExit(env, _this);

#ifdef DEBUG
  printf("MobaThread is booted.\n");
  fflush(stdout);
#endif
}


/*
 * Local Functions
 */

/*
 * ThreadRT0() (in javasrc/src/share/java/lang/thread.c)
 * $B$N(B MobaThread $BHG(B
 */
static void MobaThreadRT(Hjava_lang_Thread *thread) {
  ExecEnv ee;
  JavaFrame *frame, *prev_frame;
  struct methodblock *method;
  unsigned char *pc = NULL;

  /* $BNc30>pJs(B */
  int exc_occurred = 0;
	/* $B??56$r(B exceptionOccurred(&ee) $B$N7k2L$HF1$8$KJ]$D(B */
  JHandle *exc = NULL;


  ExecEnv *givenEE = (ExecEnv *)unhand(thread)->eetop;
	/* ThreadRT0() $B$H$O0[$J$j(B $B$9$G$K:n$C$F$"$k$b$N$rMxMQ$9$k(B */
  if (givenEE) {
    /* Thread#eetop $B$r$3$N4X?t(B (MobaThreadRT()) $B$N(B auto $BJQ?t(B ee $B$K$9$2$+$($k(B */
    memcpy(&ee, givenEE, sizeof(ExecEnv));
    unhand(thread)->eetop = (int32_t)&ee;

    free(givenEE);
	/* $BMW(B free(): readContextFrom() $B$G(B malloc() $B$5$l$F$$$k(B */
  }
  else {
    InitializeExecEnv(&ee, (JHandle *)thread);
  }


  frame = ee.current_frame;
  method = frame->current_method;


  threadInit(thread, (stackp_t) &thread);
  if (unhand(thread)->stillborn)
    threadExit();

  exceptionClear(&ee);


  /* Thread#suspend0 $B$+$i(B return $B$7$F<B9T:F3+(B */
  pc = returnPC(frame);


#ifdef DEBUG
  {
#define RT_PC_BUFSIZE	50
    char buf[RT_PC_BUFSIZE];
    pc2string(pc, method, buf, buf + RT_PC_BUFSIZE);
    printf("pc: ");
    outPC(stdout, pc, method);
    printf("\n");
    printf("MobaThreadRT(): %s\n", buf);

    printf("\n");
    fflush(stdout);
  }
#endif

  while (1) {
    bool_t execRet;

#if 0
    {
      /* Once I suspected
	 threadCheckStack() (in ExecuteJava()) returns false. */
      int ret;
      ret = sysThreadCheckStack();
      printf("threadCheckStack() returns: %s\n", (ret ? "true":"false"));
      fflush(stdout);
    }
#endif

#ifdef DEBUG
    printf("call ExecuteJava(");
    outPC(stdout, pc, frame->current_method);
    printf(")\n");
    fflush(stdout);
#endif
    {
      /* fake for JDK profiling */
      int isFrameInitial = !frame->prev;
      JavaFrame tempFrame;

      if (isFrameInitial) {
	memset(&tempFrame, 0, sizeof(JavaFrame));
	frame->prev = &tempFrame;
      }
      /* fake end */

      /* for profiling */
      if (java_monitor)  frame->mon_starttime = now();

      execRet = ExecuteJava(pc, &ee);
	/* I hope this function not to return but ... */
	/* ExecuteJava() distinguishes the initial frame and return */

      /* fake for JDK profiling */
      if (isFrameInitial)  frame->prev = NULL;
      /* fake end */
    }
#ifdef DEBUG
    printf("ExecuteJava() returns: %s\n", (execRet ? "TRUE":"FALSE"));
    fflush(stdout);
#endif

    /*exc_occurred = exceptionOccurred(&ee);*/
    exc_occurred = !execRet;	/* This way takes more low cost */
    exc = ee.exception.exc;

    if (exc_occurred) {	/* an exception occurred */
      unsigned char *new_pc;

#ifdef DEBUG
      printf("in MobaThreadRT(), exception(kind: %d) occurred:\n",
		ee.exceptionKind);
      fflush(stdout);

      if (ee.exceptionKind == EXCKIND_THROW) {
	JHandle *exc = ee.exception.exc;
	printf("  Exception: ");
	if (exc)
	  printf("%s\n", cbName(obj_classblock(exc)));
	else
	  printf("(null)\n");
      }
      exceptionDescribe(&ee);
      printf("\n");
      fflush(stderr);  fflush(stdout);
#endif

    handle_exception:
      pc = frame->lastpc;
      new_pc = ProcedureFindThrowTag(&ee, frame, exc, pc);
      if (new_pc) {	/* exception handler is found */
#ifdef DEBUG
	printf("Exception handler is found (PC: 0x%x).\n", (int)new_pc);
	fflush(stdout);
#endif
	exceptionClear(&ee);  exc_occurred = 0;
	frame->optop = frame->ostack;	/* reset to empty */

	/* push exception object */
	frame->optop->h = exc;
	frame->optop++;

	pc = frame->lastpc = new_pc;

	continue;
      }
      else {		/* exception handler is not found */
#ifdef DEBUG
	printf("Exception handler is *not* found.\n");
	fflush(stdout);
#endif
	/* In case finding the exception handler,
	   have to do the followings by myself. */
	if (frame->monitor)
	  monitorExit(obj_monitor(frame->monitor));
	if (java_monitor) {
	  struct methodblock *mb = frame->current_method;
	  if (mb && frame->prev) {
		/* assume initial_frame->prev == NULL */
	    int time = frame->mon_starttime;
	    time = ((time != 0) ? (now() - time) : 0);
		// $B%a%=%C%I(B or ExecuteJava() $B$N8F$S=P$7A0$K(B
		// frame->mon_starttime $B$,=i4|2=$5$l$F$$$J$$>l9g$r9MN8(B
#ifdef DEBUG
	    {
	      struct methodblock *caller = frame->prev->current_method;
	      struct methodblock *callee = mb;
	      printf("call java_mon(caller:%s, callee:%s, time:%d)\n",
		(caller ? (caller->fb.name):"(null)"),
		(mb ? (mb->fb.name):"(null)"), time);
	      fflush(stdout);
	    }
#endif
	    java_mon(frame->prev->current_method, mb, time);
	  }	/* if (mb && frame->prev) */
	}	/* if (java_monitor) */

	/* fall through */
      }
    }	/* if (exc_occurred) */


    /* processes should be done by the `opc_return' instruction */
    prev_frame = ee.current_frame->prev;

    /* execution finishes */
    if (!prev_frame)  break;
	/* assume initial_frame->prev == NULL */
    if (!(pc = returnPC(prev_frame)))  break;
#ifdef DEBUG
    printf("MobaThreadRT(): frame: 0x%x -> 0x%x.\n",
		(int)frame, (int)prev_frame);
    fflush(stdout);
#endif
    if (exc_occurred) {	/* there is an exception which is not processed yet */
      frame = ee.current_frame = prev_frame;
      goto handle_exception;
    }
#if EXECUTEJAVA_IN_ASM > 0
    {	/* stack a returned value gotten from the previous frame */
      int retsize;
      char *sig;
      method = frame->current_method;
      sig = method->fb.signature;
      while (*(sig++) != ')');
      switch (*sig) {
      case 'V':
	break;
      case 'L':  case '[':
      case 'Z':  case 'I':  case 'F':
	prev_frame->optop[0] = frame->optop[-1];
	prev_frame->optop++;
	break;
      case 'J':  case 'D':
	prev_frame->optop[0] = frame->optop[-2];
	prev_frame->optop[1] = frame->optop[-1];
	prev_frame->optop += 2;
	break;
      default:
#ifdef DEBUG
	printf("MobaThreadRT(): invalid return type: %c\n", *sig);
	fflush(stdout);
#endif
	break;
      }
    }
#endif	/* EXECUTEJAVA_IN_ASM */
#ifdef DEBUG
    printf("now returnd to previous frame.\n\n");
    fflush(stdout);
#endif
    frame = ee.current_frame = prev_frame;
  }	/* while (1) */


#ifdef DEBUG
  printf("MobaThreadRT(): Execution done.\n");
  fflush(stdout);
#endif

  if (exc_occurred) {	/* $BL$=hM}$NNc30$,$"$k(B */
#ifdef DEBUG
    printf("An exception is uncauted.\n");
    /* exceptionDescribe(&ee); */
    printf("  parent: 0x%x\n", (int)(unhand(thread)->group));
    fflush(stdout);
#endif
    if (unhand(thread)->group) {
      exceptionClear(&ee);
      execute_java_dynamic_method(&ee, (void *)(unhand(thread)->group),
	"uncaughtException", "(Ljava/lang/Thread;Ljava/lang/Throwable;)V",
	thread, exc);
    }
  }

  unhand(thread)->stillborn = 1;	/* true */
  /* done. */

#ifdef DEBUG
  printf("Now call Thread#exit()\n");  fflush(stdout);
#endif
  {
    int count = 3;
    while ((unhand(thread)->group) && (count-- > 0)) {
      exceptionClear(&ee);
    execute_java_dynamic_method(&ee, (void *)thread, "exit", "()V");
    }
    /* sysAssert(unhand(thread)->group == NULL); */
  }

#ifdef DEBUG
  printf("Now call threadExit()\n");  fflush(stdout);
#endif
  threadExit();
}


/*
 * opcode "return" $B$r<+A0$G=hM}$9$k!#(B
 * $B8F$V;~E@$G!"(Bframe $B$O(B return $B@h$N$b$N$K$7$F$*$/!#(B
 *
 * ExecuteJava() in javasrc/src/share/java/runtime/executeJava.c $B$K(B
 * auto $BJQ?t(B initial_frame $B$,$"$k$;$$$G!"(Bopcode "return " $B$G(B
 * ExecuteJava() $B$+$i(B return $B$7$F$7$^$&!#$J$N$G(B return $B$r<+A0$G=hM}!#(B
 *
 * returns: PC
 */
static unsigned char *returnPC(JavaFrame *frame) {
  if (frame->returnpc == 0) {
    struct methodblock *method = frame->current_method;
    struct methodblock *targetMethod =
	(struct methodblock *)frame->constant_pool;
    unsigned char *lastpc = frame->lastpc;

    if (!method) {	/* frame->current_method == NULL $BBP:v(B */
      if (lastpc)
	method = pc2method(lastpc);
      else
	return NULL;
    }

    frame->returnpc = lastpc + 3;
	/* frame->returnpc = nextPC(lastpc); */
	/* is not good, because quick insn. cause problems. */
    frame->constant_pool = cbConstantPool(fieldclass(&method->fb));
#ifdef DEBUG
    if (method && lastpc && frame->returnpc) {
      uint32_t off0, off1;
      off0 = pc2offset(method, lastpc);
      off1 = pc2offset(method, frame->returnpc);
      printf("pc(offset): %u -> %u  (in returnPC())\n", off0, off1);
      fflush(stdout);
    }
#endif

#if 0	/* comment func. call FixupQuickInvocation() out. */
#if (JDK_VER >= 110)
#ifdef DEBUG
    printf("call FixupquickInvoction():\n");
    printf(" lastpc: ");  outPC(stdout, lastpc, method);  printf("\n");
    printf(" current_method: 0x%x ", (int)method);
    outMethod(stdout, method);
    printf("\n");
    printf(" target_method:  0x%x ", (int)targetMethod);
    outMethod(stdout, targetMethod);
    printf("\n");
    fflush(stdout);
#endif
#ifdef DEBUG
    printf("before FixupQuickInvocation(): *lastpc: %3d %3d %3d %3d %3d\n",
	(int)*lastpc, (int)*(lastpc+1), (int)*(lastpc+2),
	(int)*(lastpc+3), (int)*(lastpc+4));
#endif
    FixupQuickInvocation(lastpc, method, targetMethod);
#ifdef DEBUG
    printf("after  FixupQuickInvocation(): *lastpc: %3d %3d %3d %3d %3d\n",
	(int)*lastpc, (int)*(lastpc+1), (int)*(lastpc+2),
	(int)*(lastpc+3), (int)*(lastpc+4));
#endif
#endif	/* JDK_VER >= 110 */
#endif	/* comment func. call FixupQuickInvocation() out. */
  }

#if 0
  printf("returnPC() returns: 0x%x\n", (int)frame->returnpc);
  fflush(stdout);
#endif
  return frame->returnpc;
}


/*
 * create a new JavaFrame on the stack.
 * cf. invokeJavaMethod() in javasrc/src/share/java/runtime/classruntime.c
 */
static JavaFrame *allocNewJavaFrame(struct methodblock *mb, ExecEnv *ee) {
  int nlocals = (mb ? mb->nlocals : 0);
#if 0	/* the method to calculate nlocals by myself */
  mb->nlocals = nlocals =
	(mb->fb.access & ACC_NATIVE)?0:
		(mb->args_size + ((mb->fb.access & ACC_STATIC)?0:1));
#endif

  return allocNewJavaFrameWithNLocals(mb, ee, nlocals);
}


/*
 * create a new JavaFrame on the stack.
 * cf. invokeJavaMethod() in javasrc/src/share/java/runtime/classruntime.c
 */
static JavaFrame *allocNewJavaFrameWithNLocals(struct methodblock *mb, ExecEnv *ee, int nlocals) {
  JavaFrame *prev_frame;
  stack_item *prev_optop;
  JavaStack *stack;
  JavaFrame *frame;

#ifdef DEBUG
    printf("allocNewJavaFrame((methodblock *)0x%x, (ExecEnv *)0x%x) is called.\n", (int)mb, (int)ee);
    if (mb)  printf(" method name: %s\n", mb->fb.name);
    fflush(stdout);
#endif

  if (!ee->current_frame) {	/* $B$R$H$D$a$N(B JavaFrame */
#ifdef DEBUG
    printf(" first frame.\n");  fflush(stdout);
#endif
    prev_frame = NULL;
#ifdef DEBUG
    if (!ee->initial_stack) {
      /* not reached */
      printf(" ee->initial_stack is NULL!\n");  fflush(stdout);
    }
#endif
    prev_optop = ee->initial_stack->data;
    stack = ee->initial_stack;
  }
  else {
    prev_frame = ee->current_frame;
    prev_optop = prev_frame->optop;
    stack = prev_frame->javastack;	/* current stack chunk */
  }
#ifdef DEBUG
  printf(" stack:       0x%x\n", (int)stack);
  if (stack)
    printf(" stack->next: 0x%x\n", (int)stack->next);
  fflush(stdout);
#endif

  frame = (JavaFrame *)(prev_optop + nlocals);

  if (frame->ostack + (mb ? mb->maxstack : 0) >= stack->end_data) {
    /* current stack chunk $B$K?7(B frame $B$,<}$^$i$J$$(B */
#ifdef DEBUG
    printf(" new frame doesn't fall into current stack chunk.\n");
    fflush(stdout);
#endif
    if (stack->next != 0) {
#ifdef DEBUG
      printf(" use next stack chunk.\n");
      fflush(stdout);
#endif
      stack = stack->next;
    }
    else {
      /* $B?7$7$$(B stack chunk $B$r3NJ](B */
#ifdef DEBUG
    printf(" must allocate new stack chunk.\n");
    fflush(stdout);
#endif
      if (stack->stack_so_far +
		JAVASTACK_CHUNK_SIZE*sizeof(stack->data[0]) > JavaStackSize) {
	SignalError(EE(), "java/lang/StackOverflowError", 0);
		/* $B$3$N;~E@$N(B ExecEnv $B$rF@$k$?$a$K!"4:$($F(B ee $B$G$O$J$/(B EE() */
	return FALSE;
      }

      stack = CreateNewJavaStack(ee, stack);
      if (stack == 0) {
	SignalError(ee, "java/lang/OutOfMemoryError", 0);  return FALSE;
      }
    }

    /* $B?7(B stack chunk $B>e$K!"?7(B frame $B$N>l=j$rDj$a$k(B */
    frame = (JavaFrame *)(stack->data + nlocals);
    if (frame->ostack + (mb ? mb->maxstack : 0) >= stack->end_data) {
      /* stack chunk $B$K(B frame $B$,<}$^$i$J$$(B */
      SignalError(EE(), "java/lang/InternalError", 0);  return FALSE;
    }

    prev_optop = stack->data;
  }	/* if (stack chunk $B$K(B frame $B$,<}$^$i$J$$(B) */
#ifdef DEBUG
    printf(" frame: 0x%x\n", (int)frame);
    fflush(stdout);
#endif


  /* configure members of JavaFrame */
  frame->javastack = stack;
  frame->prev = prev_frame;
  frame->vars = prev_optop;
  frame->optop = frame->ostack;

  frame->current_method = mb;
  frame->constant_pool = (mb ? cbConstantPool(fieldclass(&mb->fb)) : NULL);
  frame->returnpc = (mb ? mb->code : 0);
  frame->monitor = 0;
  ee->current_frame = frame;

  frame->mon_starttime = 0;
  /* if (java_monitor) { frame->mon_starttime = now(); } */

#ifdef DEBUG
  printf(" constant_pool: 0x%x\n", (int)frame->constant_pool);

  printf("allocNewJavaFrame() done.\n");
  fflush(stdout);
#endif

  return frame;
}


/*
 * struct methodblock $B$r=q$-=P$9!#(B
 */
static int writeClassAndMethod(JNIEnv *env,
		jobject /*OutputStream*/ out, struct methodblock *method) {
  ClassClass *clazz;
  char *name, *sig;

  clazz = method->fb.clazz;
  name = method->fb.name;
  sig = method->fb.signature;

#ifdef DEBUG
  printf("writeClassAndMethod():\n");
  printf(" class name: %s\n", cbName(clazz));
  printf(" method: %s %s\n", name, sig);
  fflush(stdout);
#endif
  if ((writeString(env, out, cbName(clazz))) < 0)  return -1;/* $B%/%i%9L>(B */
  if ((writeString(env, out, name)) < 0)  return -1;	/* $B%a%=%C%IL>(B */
  if ((writeString(env, out, sig)) < 0)  return -1;	/* signature */
#if 0
  printf("writeClassAndMethod() done.\n");
  fflush(stdout);
#endif

  return 0;
}


/*
 * struct methodblock $B$rFI$_9~$`!#(B
 */
static struct methodblock *readClassAndMethod(
	JNIEnv *env, jobject /*InputStream*/ in, jobject loader) {
  char *className, *name, *sig;
  jclass clazz;
  jmethodID method;

  jstring classNameStr;

#ifdef DEBUG
printf("readClassAndMethod() called.\n");  fflush(stdout);
#endif

  className = readStrdup(env, in);	/* $B%/%i%9L>(B */
  name = readStrdup(env, in);		/* $B%a%=%C%IL>(B */
  sig = readStrdup(env, in);		/* signature */
#ifdef DEBUG
  printf("class name: %s\n", className);
  printf("method: %s %s\n", name, sig);
  fflush(stdout);
#endif

  clazz = classLoaderFindClass(env, className, loader);
  if ((*env)->ExceptionOccurred(env)) {
    /* not reached */
    printf("readClassAndMethod(): class %s is not found.\n", className);
    fflush(stdout);
    return NULL;
  }
#ifdef DEBUG
  printf("readClassAndMethod(): find class returns: 0x%x\n", (int)clazz);
  fflush(stdout);
#endif

  method = (*env)->GetMethodID(env, clazz, name, sig);
  if ((*env)->ExceptionOccurred(env)) {
    (*env)->ExceptionClear(env);
    method = (*env)->GetStaticMethodID(env, clazz, name, sig);
    if ((*env)->ExceptionOccurred(env)) {
      /* not reached */
      printf("readClassAndMethod(): method %s is not found.\n", name);
      fflush(stdout);
      return NULL;
    }
  }

  free(className);
  free(name);
  free(sig);

  return (struct methodblock *)method;
}


/*
 * stack_item $B$r=q$-=P$9!#(B
 * $B%*%V%8%'%/%H$+B(CM$+$O!"0z?t$N(B signature $B$GH=Dj$9$k!#(B
 */
static int writeStackItem(JNIEnv *env,
		jobject /*ObjectWriter*/ out, stack_item *itemptr, char *sig) {
#ifdef DEBUG
  printf("writeStackItem((stack_item)0x%x, sig \"%s\") called.\n", *((int *)itemptr), sig);
  fflush(stdout);
#endif
  if ((sig[0] == 'L') || (sig[0] == '[')) {
#ifdef DEBUG
    printf("stack item is handle.\n");
    fflush(stdout);
#endif
    if ((writeChar(env, out, STACK_ITEM_HANDLE)) < 0)
      return -1;
    if ((writeObject(env, out,
		MkRefLocal(env, /*item.h $B$G$O$J$/(B*/ *((JHandle **)itemptr),
			JNI_REF_HANDLE_TAG))) < 0)
      return -1;
	/* serialize */
  }
  else {
    if ((sig[0] == SIGNATURE_DOUBLE) || (sig[0] == SIGNATURE_LONG)) {
      Java8 tmp64;
      GET_INT64(tmp64, itemptr);
#ifdef DEBUG
      printf("stack item is 64bit immediate value: 0x%08x, 0x%08x\n",
		itemptr->i, (itemptr + 1)->i);
      printf(" substitute to: 0x%08x, 0x%08x\n",
		tmp64.x[0], tmp64.x[1]);
      fflush(stdout);
#endif
      if ((writeChar(env, out, STACK_ITEM_IMM64)) < 0)
	return -1;
      if ((writeInt64(env, out, (jlong *)&tmp64)) < 0)
	return -1;
    }
    else {
#ifdef DEBUG
    printf("stack item is 32bit immediate value: %d\n", itemptr->i);
    fflush(stdout);
#endif
      if ((writeChar(env, out, STACK_ITEM_IMM32)) < 0)
	return -1;
      if ((writeInt32(env, out, (jint *)itemptr)) < 0)
	return -1;
    }
  }

  return 0;
}

/*
 * stack_item $B$r=q$-=P$9!#(B
 * $B%*%V%8%'%/%H$+B(CM$+$r(B signature $B$KMj$i$:FbIt$GH=Dj$9$k!#(B
 */
static int writeStackItemAuto(JNIEnv *env,
		jobject /*ObjectWriter*/ out, stack_item *itemptr) {
#ifdef DEBUG
  printf("writeStackItemAuto(0x%x) called.\n", *((int *)itemptr));
  fflush(stdout);
#endif
  if (isPtrHandle((void *)itemptr->h)) {
#ifdef DEBUG
    printf("stack item is handle: ");
    if (itemptr->h) {
      if (obj_flags(itemptr->h) == T_NORMAL_OBJECT)
	printf("%s", cbName(obj_classblock(itemptr->h)));
      else
	printf("array");
    }
    else
      printf("(null)");
    printf("\n");
    fflush(stdout);
#endif
    if ((writeChar(env, out, STACK_ITEM_HANDLE)) < 0)
      return -1;
    if ((writeObject(env, out,
	MkRefLocal(env, itemptr->h, JNI_REF_HANDLE_TAG))) < 0)  return -1;
	/* serialize */
  }
  else {
#ifdef DEBUG
    printf("stack item is immediate value.\n");
    fflush(stdout);
#endif
    if ((writeChar(env, out, STACK_ITEM_IMM32)) < 0)
      return -1;
    if ((writeInt32(env, out, ((jint *)itemptr))) < 0)
      return -1;
  }

#ifdef DEBUG
  printf("writeStackItemAuto() done.\n");
  fflush(stdout);
#endif
  return 0;
}

/*
 * stack_item $B$rFI$_9~$`!#(B
 *
 * returns: item type (STACK_ITEM_*)
 */
static int readStackItem(JNIEnv *env,
		jobject/*ObjectReader*/ in, stack_item *dest) {
  char itemType = -1;
#ifdef DEBUG
    printf("readStackItem() called.\n");
    fflush(stdout);
#endif

  itemType = readChar(env, in);
#ifdef DEBUG
    printf("item type is %d.\n", (int)itemType);
    fflush(stdout);
#endif

  switch ((int)itemType) {
  case STACK_ITEM_HANDLE:
#ifdef DEBUG
    printf("stack item is handle.\n");
    fflush(stdout);
#endif
    {
      jobject obj = readObject(env, in);	/* de-serialize */
      dest->h = (JHandle *)DeRef(env, obj);
    }
#ifdef DEBUG
    printf("readStackItem(): readObject() returns 0x%x\n", (int)dest);
    fflush(stdout);
#endif
    break;
  case STACK_ITEM_IMM32:
    dest->i = readInt32(env, in);
#ifdef DEBUG
    printf("stack item is 32bit immediate value: %d\n", dest->i);
    fflush(stdout);
#endif
    break;
  case STACK_ITEM_IMM64:
    {
      jlong val;
      Java8 tmp64;
      val = readInt64(env, in);
      SET_INT64(tmp64, dest, val);	/* *((jlong *)&dest) = val */
#ifdef DEBUG
      printf("stack item is 64bit immediate value: 0x%08x, 0x%08x\n",
		dest->i, (dest + 1)->i);
      printf(" substitute from: 0x%08x, 0x%08x\n",
		*((jint *)&val), *(((jint *)&val) + 1));
      fflush(stdout);
#endif
    }
    break;
  default:
#ifdef DEBUG
    printf("stream data is invalid...: 0x%X\n", (int)itemType & 0xff);
    fflush(stdout);
#endif
    SignalError(0, MOBA_PACKAGE "StreamFormatException",
		"STACK_ITEM_* is expected.");
    return;
    break;
  }
#ifdef DEBUG
  printf("readStackItem() done.\n");
  fflush(stdout);
#endif

  return (int)itemType;
}


/*
 * JavaFrame $B$NCf?H$r%@%s%W$9$k!#(B
 * for debug
 */
static void dumpAllFrames(FILE *out, JavaFrame *topFrame) {
  JavaFrame *frame;
  int count;

  fprintf(out, "Now dump all frames:\n");
  fprintf(out, "\n");

  /* stack frame $B?t$r?t$($k(B */
  frame = topFrame;
  count = 0;
  while (frame) {
    count++;
    frame = frame->prev;
  }

  frame = topFrame;
  while (frame) {
    fprintf(out, "dumping stack frame(%d):\n", --count);
    dumpFrame(out, frame);
    printf("\n");

    frame = frame->prev;
  }
}


static void dumpFrame(FILE *out, JavaFrame *frame) {
  struct methodblock *method = NULL;
  ClassClass *clazz = NULL;
  int isReturnpcZero = 0;	/* false */

  fprintf(out, "address of frame:\t0x%x\n", (int)frame);
  fflush(out);

  if (!frame)  return;

  method = frame->current_method;
  if (method)  clazz = method->fb.clazz;

  /* returnpc */
  if (!frame->returnpc)  isReturnpcZero = 1;	/* true */

  fprintf(out, "returnpc:\t\t");
  outPC(out, frame->returnpc, frame->current_method);  fprintf(out, "\n");
  fflush(out);

  /* constant_pool */
  /* $B7?$,(B returnpc $B$K0MB8$9$k(B
   *  returnpc is 0: (struct methodblock *)
   *       is not 0: (cp_itemm *)
   */
  {
    fprintf(out, "constant_pool:\t\t0x%x", (int)frame->constant_pool);
    fflush(out);

    if (!frame->constant_pool) {
      fprintf(out, " (0 is illegal!)\n");
    }
    else {
      if (isReturnpcZero) {
	struct methodblock *targetMethod =
	  (struct methodblock *)frame->constant_pool;
	fprintf(out, " (method: %s#%s)",
		cbName(targetMethod->fb.clazz), targetMethod->fb.name);
      }
      else {
	ClassClass *cpClazz = cp2class(frame->constant_pool);
	int index;
	if (cpClazz) {
	  /* CP is resolved. */
	  cp_item_type *cbCP = cbConstantPool(cpClazz);
	  index = (int)cp2offset(cpClazz, frame->constant_pool);
	  fprintf(out, " (offset: %d) in %s", index, cbName(cpClazz));
	}
	else {
	  index = frame->constant_pool->i;
	  fprintf(out, " (offset: %d)", index);
	}
      }
    }

    fprintf(out, "\n");
    fflush(out);
  }

  /* optop */
  {
    int optopOffset;
    optopOffset = (int)sp2offset(frame, frame->optop);
    fprintf(out, "optop:\t\t\t0x%x (offset: %d)\n",
	(int)frame->optop, optopOffset);
    fflush(out);
  }

  /* vars */
  {
    int varsOffset;
    varsOffset = (int)sp2offset(frame, frame->vars);
    fprintf(out, "vars:\t\t\t0x%x (offset: %d)\n",
	(int)frame->vars, varsOffset);
    fflush(out);
  }

  /* prev */
  fprintf(out, "prev:\t\t\t0x%x\n", (int)frame->prev);
  /* javastack */
  fprintf(out, "javastack:\t\t0x%x\n", (int)frame->javastack);

  /* lastpc */
  fprintf(out, "lastpc:\t\t\t");
  outPC(out, frame->lastpc, frame->current_method);  fprintf(out, "\n");
  fflush(out);

  /* current_method */
  fprintf(out, "current_method:\t\t0x%x (%s#%s)\n",
	  (int)method,
	  (method ? (cbName(method->fb.clazz)) : "null"),
	  (method ? (method->fb.name) : "null"));
  fflush(out);

  /* OTHERS */
  fprintf(out, "monitor:\t\t0x%x\n", (int)frame->monitor);
  fprintf(out, "mon_starttime:\t\t%d\n", frame->mon_starttime);
  fprintf(out, "ostack:\t\t\t0x%x\n", (int)frame->ostack);
  /* fprintf(out, ":\t0x%x\n", frame->); */

  fflush(out);

  return;
}
