/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

package NET.shudo.moba.place;


import java.net.Socket;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.util.Dictionary;


/**
 * MobaClassLoader ׵˱
 * 饹(饹ե)ͥåȥۤۤ륯饹<BR>
 * ClassServer 鵯ư롣
 *
 * @see NET.shudo.moba.place.ClassServer
 * @see NET.shudo.moba.place.MobaClassLoader
 */
public
class ClassDistributor implements Runnable {
  Socket sock = null;
  Dictionary classTable = null;
  boolean debug = false;

  InputStream sockIn = null;
  OutputStream sockOut = null;
  DataInputStream in = null;
  DataOutputStream out = null;


  protected ClassDistributor(Socket sock, Dictionary classTable, boolean debug)
	throws IOException {
    this.sock = sock;
    this.classTable = classTable;
    this.debug = debug;

    // 饹̾դ
    try {
      sockIn = sock.getInputStream();
      sockOut = sock.getOutputStream();
    }
    catch (IOException e) {
      if (debug)  e.printStackTrace();
    }
    in = new DataInputStream(new BufferedInputStream(sockIn));
    out = new DataOutputStream(new BufferedOutputStream(sockOut));
  }
  protected ClassDistributor(Socket sock, Dictionary classTable)
	throws IOException {
    this(sock, classTable, false);
  }


  public void run() {
    while (true) {
      // 饹õ reply
      String classname = null;
      byte[] classfile = null;

      try {
	classname = in.readUTF();	// 饹̾
	if (debug) {
	  System.out.println("ClassDistributor#run() classname: " + classname);
	  System.out.flush();
	}

	synchronized(classTable) {
	  classfile = (byte[])classTable.get(classname);
	}
	if (classfile == null) {
	  classfile = LocalClassLoader.loadClassfileLocally(classname);
	  if (classfile == null || classfile.length <= 0) {
	    out.writeInt(-1);
	    out.flush();
	    continue;
	  }

	  synchronized(classTable) {
	    classTable.put(classname, classfile);
	  }
	}

	if (debug) {
	  System.out.println("ClassDistributor#run() length: " +
		classfile.length);
	  System.out.flush();
	}
	out.writeInt(classfile.length);
	out.flush();
	out.write(classfile);
	out.flush();
      }
      catch (IOException e) {
	if (debug) {
	  System.out.println("Now this ClassDistributor will be freed.");
	  System.out.println("Don't worry, this exception is inevitable one: ");
	  e.printStackTrace();
	}
	break;
      }
    }	// while (true)
  }
}
